import pandas as pd
import numpy as np
import datetime
import json

import nltk
nltk.data.path.append('/var/task/nltk_data/')
from nltk.tokenize import word_tokenize
from nltk.corpus import stopwords
s=set(stopwords.words('english'))


from textblob import TextBlob
from DataScience import DataScience
from config import datascience_format_options


# flatten a list of lists into a list
flatten = lambda l: [item for sublist in l for item in sublist]

# convert raw text into sentences
text_to_sentences = lambda text: flatten(map(nltk.sent_tokenize, filter(lambda x : x != '', text.split('\n'))))

# convert raw text to tokenized report of words and punctuation
text_to_tokenized_reports = lambda text: [word_tokenize(sentence) for sentence in text_to_sentences(text)] 

# convert raw text to list of list of nouns
text_to_noun_sentences = lambda text: [list(TextBlob(sentence).noun_phrases) for sentence in text_to_sentences(text)]

# convert raw text to list of nouns
text_to_nouns = lambda text: flatten(text_to_noun_sentences(text))

# convert raw text to list of words
text_to_words = lambda text: [x.lower() for x in flatten(text_to_tokenized_reports(text)) if x.isalpha()]
text_to_words_rm_stop = lambda text: [x.lower() for x in filter(lambda w: not w in s, flatten(text_to_tokenized_reports(text))) if x.isalpha()]


# gets all of the words in a series of a list of words
get_all_words = lambda series: list(set(flatten(list(series.values))))

def split_to_discrete_word_matrix(df, now = None, indexes = ['created_time'], keys = ['name'], name = None, func = text_to_words):
    if now is None:
        now = datetime.datetime.now()

    df = df.copy()
    
    found = False
    for ind in indexes:
        if ind in df:
            df[ind] = df[ind].bfill()
            df = df.set_index(ind)
            found = True
            break
    if not found:
        df.index = [now for _ in list(df.index)]
            
    all_words = set()

    good_keys = [key + '__words' for key in keys]
    for key in keys:
        df[key + '__words'] = df[key].fillna('').apply(func)
        all_words = all_words.union(set(get_all_words(df[key + '__words'])))

    all_words = map(unicode, sorted(list(all_words)))
    return df[good_keys], all_words

def data_science_to_text_processor(data, name):
    return DataScienceToTextProcessor(data, name, func = text_to_words_rm_stop).to_json()   

class DataScienceToTextProcessor(DataScience):
    def __init__(self, data, name, func = text_to_words):
        super(DataScienceToTextProcessor, self).__init__(data, name)
        self.func = func
        self.text_process()
    
    def to_string(self):
        base = super(DataScienceToTextProcessor, self).to_string()
        base += '--------- TEXT PROCESSOR ---------\n'
        for attr in self.attributes:
            base += '|- %s__words\n' % attr
            base += '|- %s__allwords\n' % attr
        return base
    
    def text_process(self):
        valid_text = ['raw_string', 'raw_text', 'raw_enum', 'raw_username_string']
        text_keys = set(filter(lambda x : self.col_to_type[x] in valid_text, self.col_to_type.keys()))
        def build_params(attr):
            df =  getattr(self, attr)
            df.columns
            return {
                'df' : df,
                'name' : '%s.%s' % (self.name, attr),
                'keys' : list(filter(lambda x : x in text_keys, list(df.columns))),
                'func' : self.func,
                'indexes' : datascience_format_options[self.name]['indexes'],
                'now' : self.now
                
            }
        for attr in self.attributes:
            params = build_params(attr)
            df, all_words = split_to_discrete_word_matrix(**params)
            setattr(self, '%s__words' % attr, df)
            setattr(self, '%s__allwords' % attr, all_words)
    
    def to_json(self):
        base = super(DataScienceToTextProcessor, self).to_json()
        base['textProcesserAttributes'] = {}
        for attr in self.attributes:
            df = getattr(self, '%s__words' % attr)
            all_words = getattr(self, '%s__allwords' % attr)
            df['__index__'] = df.index
            base['textProcesserAttributes']['%s__words' % attr] = json.loads(df.to_json(orient='records'))
            base['textProcesserAttributes']['%s__allwords' % attr] = all_words
        return base

class TextProcessor(DataScience):
    def __init__(self, data, name, alpha = 0):
        super(TextProcessor, self).__init__(data, name)
        self.alpha = alpha
        self.textProcesserAttributes = self.allData['textProcesserAttributes'].keys()
        self.expand_to_sparse()

    def expand_to_sparse(self):

        for attr in self.attributes:
            df =  self.to_df(self.allData['textProcesserAttributes']['%s__words' % attr], attr, index = '__index__')
            all_words = self.allData['textProcesserAttributes']['%s__allwords' % attr]
            counts_df = pd.DataFrame(self.alpha, index = df.index, columns = all_words)
            full_df = pd.concat([df, counts_df], axis = 1)

            keys = list(df.columns)
            weights = [1 for _ in keys]
            zipped = zip(keys, weights)
            def word_list_to_counts(row):
                for key, weight in zipped:
                    for word in row[key]:
                        try:
                            row[word] += weight
                        except:
                            print word
                return row
            full_df = full_df.apply(word_list_to_counts, axis = 1)
            setattr(self, '%s__words' % attr, full_df.ix[:, len(keys):])
    
    def to_string(self):
        base = super(TextProcessor, self).to_string()
        base += '--------- TEXT PROCESSOR ---------\n'
        for attr in self.attributes:
            base += '|- %s__words\n' % attr
        return base
    
    def to_json(self):
        base = super(TextProcessor, self).to_json()
        base['textProcesserAttributes'] = {}
        for attr in self.attributes:
            base['textProcesserAttributes'][attr] = getattr(self, '%s__words' % attr).to_json(orient='records')
        return base




datascience_format_options = {
    'spotify' : {
        'attributes' : ['savedTracks', 'followedArtists', 'playHistory', 'savedPlaylists', 'profile'],
        'col_to_type' : {
            'addedAt' : 'raw_datetime',
            'album' : 'raw_string',
            'artists' : 'raw_list',
            'birthdate' : 'raw_date',
            'context' : 'raw_object',
            'context.external_urls.spotify' : 'url',
            'context.href' : 'raw_url',
            'context.type' : 'raw_enum',
            'context.uri' : 'raw_url',
            'country' : 'raw_string',
            'createdAt' : 'raw_datetime',
            'display_name' : 'raw_string',
            'email' : 'raw_email',
            'external_urls.spotify' : 'raw_url',
            'followers' : 'raw_integer',
            'followers.href' : 'raw_url',
            'followers.total' : 'raw_integer',
            'genres' : 'raw_list',
            'href' : 'raw_url',
            'id' : 'id',
            'images' : 'raw_list',
            'img' : 'raw_url',
            'latestPull' : 'raw_datetime',
            'metaData.acousticness' : 'raw_float',
            'metaData.analysis_url' : 'raw_url',
            'metaData.danceability' : 'raw_float',
            'metaData.duration_ms' : 'raw_integer',
            'metaData.energy' : 'raw_float',
            'metaData.id' : 'id',
            'metaData.instrumentalness' : 'raw_integer',
            'metaData.key' : 'raw_integer',
            'metaData.liveness' : 'raw_float',
            'metaData.loudness' : 'raw_float',
            'metaData.mode' : 'raw_integer',
            'metaData.speechiness' : 'raw_float',
            'metaData.tempo' : 'raw_float',
            'metaData.time_signature' : 'raw_integer',
            'metaData.track_href' : 'raw_url',
            'metaData.type' : 'raw_enum',
            'metaData.uri' : 'raw_url',
            'metaData.valence' : 'raw_float',
            'name' : 'raw_string',
            'numTracks' : 'raw_integer',
            'played_at' : 'raw_datetime',
            'popularity' : 'raw_integer',
            'previewUrl' : 'raw_url',
            'product' : 'raw_enum',
            'public' : 'raw_boolean',
            'timestamp' : 'raw_datetime',
            'track.artists' : 'raw_list',
            'track.duration_ms' : 'raw_integer',
            'track.explicit' : 'raw_boolean',
            'track.href' : 'raw_url',
            'track.id' : 'id',
            'track.name' : 'raw_string',
            'track.popularity' : 'raw_integer',
            'track.preview_url' : 'raw_url',
            'track.track_number' : 'raw_integer',
            'tracks' : 'raw_list',
            'type' : 'raw_enum',
            'updatedAt' : 'raw_datetime',
            'uri' : 'raw_url'
        },
        'date_str_strip' : '%Y-%m-%d',
        'indexes' : ['addedAt', 'timestamp', 'updatedAt', 'played_at'],
        'getAttributes' : {
            'topTracks' : 'tracks', 
            'topArtists' : 'artists'
        }
    },
    'reddit' : {
        'attributes' : ['subscribed', 'comments', 'karmaList', 'savedPosts', 'submittedPosts', 
                      'upvoted', 'downvoted', 'profile'],
        'col_to_type' : {
            'accountCreated' : 'raw_integer',
            'author' : 'raw_username_string',
            'commentKarma' : 'raw_integer',
            'created' : 'raw_datetime',
            'createdAt' : 'raw_datetime',
            'downs' : 'raw_integer',
            'id' : 'id',
            'isEmployee' : 'raw_boolean',
            'isGold' : 'raw_boolean',
            'isMod' : 'raw_boolean',
            'isSponsor' : 'raw_boolean',
            'isVideo' : 'raw_boolean',
            'kind' : 'raw_enum',
            'latestPull' : 'raw_datetime',
            'linkKarma' : 'raw_integer',
            'name' : 'raw_string',
            'numComments' : 'raw_integer',
            'numSubscribers' : 'raw_integer',
            'over18' : 'raw_boolean',
            'postBody' : 'raw_text',
            'saved' : 'raw_boolean',
            'score' : 'raw_integer',
            'subreddit' : 'raw_enum',
            'subreddit_id' : 'id',
            'title' : 'raw_text',
            'updatedAt' : 'raw_datetime',
            'ups' : 'raw_integer',
            'url' : 'raw_url',
            'username' : 'raw_username_string',
            'view_count' : 'raw_integer'
        },
        'indexes' : ['created', 'updatedAt']
    },
    'instagram' : {
        'attributes' : ['posts', 'likes', 'followers', 'following', 'counts', 'profile'],
        'col_to_type' : {
            '_filter' : 'raw_enum',
            'bio' : 'raw_text',
            'caption' : 'raw_text',
            'commentsCount' : 'raw_integer',
            'counts.followed_by' : 'raw_integer',
            'counts.follows' : 'raw_integer',
            'counts.media' : 'raw_integer',
            'createdAt' : 'raw_timestamp',
            'filter' : 'raw_enum',
            'followed_by' : 'raw_integer',
            'follows' : 'raw_integer',
            'full_name' : 'raw_string',
            'id' : 'id',
            'is_business' : 'raw_boolean',
            'latestPull' : 'raw_timestamp',
            'likesCount' : 'raw_integer',
            'link' : 'raw_url',
            'location' : 'raw_object',
            'location.id' : 'id',
            'location.latitude' : 'raw_float',
            'location.longitude' : 'raw_float',
            'location.name' : 'raw_string',
            'media' : 'raw_integer',
            'name' : 'raw_string',
            'photoURL' : 'raw_ur;',
            'postAuthorId' : 'id',
            'postAuthorName' : 'raw_string',
            'postAuthorUsername' : 'raw_username_string',
            'postType' : 'raw_enum',
            'profile_picture' : 'raw_url',
            'tags' : 'raw_list',
            'timePulled' : 'raw_datetime',
            'timestamp' : 'raw_datetime',
            'updatedAt' : 'raw_datetime',
            'username' : 'raw_username_string',
            'usersTagged' : 'raw_list',
            'website' : 'raw_url'
        },
        'indexes' : ['timestamp', 'timePulled', 'updatedAt']
    },
    'rescuetime' : {
        'attributes' : ['data'],
        'col_to_type' : {
            'activity' : 'raw_enum',
            'category' : 'raw_enum',
            'createdAt' : 'raw_timestamp',
            'date' : 'raw_timestamp',
            'latestPull' : 'raw_timestamp',
            'productivity' : 'raw_integer',
            'timeSpent' : 'raw_integer',
            'updatedAt' : 'raw_timestamp',   
        },
        'indexes' : ['date', 'updatedAt']
    },
    'youtube' : {
        'attributes' : ['uploadedVideos', 'subscriptions', 'myActivity', 'likes', 'profile'],
        'col_to_type' : {
            'accountCreated' : 'raw_datetime',
            'action.like.resourceId.kind' : 'raw_enum',
            'action.like.resourceId.videoId' : 'id',
            'action.playlistItem.playlistId' : 'id',
            'action.playlistItem.playlistItemId' : 'id',
            'action.playlistItem.resourceId.kind' : 'raw_enum',
            'action.playlistItem.resourceId.videoId' : 'id',
            'action.subscription.resourceId.channelId' : 'id',
            'action.subscription.resourceId.kind' : 'raw_enum',
            'action.upload.videoId' : 'id',
            'actionDetails.actionBy' : 'raw_string',
            'actionDetails.actionDate' : 'raw_datetime',
            'actionDetails.actionType' : 'raw_enum',
            'channelDescription' : 'raw_text',
            'channelId' : 'id',
            'channelTitle' : 'raw_string',
            'createdAt' : 'raw_datetime',
            'dateLiked' : 'raw_datetime',
            'description' : 'raw_text',
            'etag' : 'raw_url_end',
            'id' : 'id',
            'kind' : 'raw_enum',
            'latestPull' : 'raw_datetime',
            'likeId' : 'id',
            'relatedPlaylists.favorites' : 'raw_url_end',
            'relatedPlaylists.likes' : 'raw_url_end',
            'relatedPlaylists.uploads' : 'raw_url_end',
            'relatedPlaylists.watchHistory' : 'raw_url_end',
            'relatedPlaylists.watchLater' : 'raw_url_end',
            'subscribedAt' : 'raw_datetime',
            'thumbnails.default.height' : 'raw_integer',
            'thumbnails.default.url' : 'raw_url',
            'thumbnails.default.width' : 'raw_integer',
            'thumbnails.high.height' : 'raw_integer',
            'thumbnails.high.url' : 'raw_url',
            'thumbnails.high.width' : 'raw_integer',
            'thumbnails.maxres.height' : 'raw_integer',
            'thumbnails.maxres.url' : 'raw_url',
            'thumbnails.maxres.width' : 'raw_integer',
            'thumbnails.medium.height' : 'raw_integer',
            'thumbnails.medium.url' : 'raw_url',
            'thumbnails.medium.width' : 'raw_integer',
            'thumbnails.standard.height' : 'raw_integer',
            'thumbnails.standard.url' : 'raw_url',
            'thumbnails.standard.width' : 'raw_integer',
            'updatedAt' : 'raw_datetime',
            'videoDescription' : 'raw_text',
            'videoId' : 'id',
            'videoPublishedAt': 'raw_datetime',
            'videoThumbnails' : 'raw_object',
            'videoThumbnails.default.height' : 'raw_integer',
            'videoThumbnails.default.url' : 'raw_url',
            'videoThumbnails.default.width' : 'raw_integer',
            'videoThumbnails.high.height' : 'raw_integer',
            'videoThumbnails.high.url' : 'raw_url',
            'videoThumbnails.high.width' : 'raw_integer',
            'videoThumbnails.maxres.height' : 'raw_integer',
            'videoThumbnails.maxres.url' : 'raw_url',
            'videoThumbnails.maxres.width' : 'raw_integer',
            'videoThumbnails.medium.height' : 'raw_integer',
            'videoThumbnails.medium.url' : 'raw_url',
            'videoThumbnails.medium.width' : 'raw_integer',
            'videoThumbnails.standard.height' : 'raw_integer',
            'videoThumbnails.standard.url' : 'raw_url',
            'videoThumbnails.standard.width' : 'raw_integer',
            'videoTitle' : 'raw_text'
        },
        'indexes' : ['videoPublishedAt', 'subscribedAt', 'actionDetails.actionDate', 'dateLiked', 'updatedAt']
    },
    'linkedin' : {
         'additional_details' : ['formattedName', 'numConnections',  'firstName', 'headline', 
                              'lastName', 'industry', 'summary', 'siteStandardProfileRequest', 
                              'pictureUrl', 'location', 'publicProfileUrl', 'id',
                              'updatedAt', 'createdAt', 'latestPull'],
        'attributes' : ['shares', 'positions'],
        'col_to_type' : {
            'author.firstName' : 'raw_string',
            'author.id' : 'id',
            'author.lastName' : 'raw_string',
            'company.id' : 'id',
            'company.industry' : 'raw_enum',
            'company.name' : 'raw_string',
            'content.description' : 'raw_text',
            'content.eyebrowUrl' : 'raw_url',
            'content.resolvedUrl' : 'raw_url',
            'content.shortenedUrl' : 'raw_url',
            'content.submittedImageUrl' : 'raw_url',
            'content.submittedUrl' : 'raw_url',
            'content.thumbnailUrl' : 'raw_url',
            'content.title' : 'raw_text',
            'createdAt' : 'raw_datetime',
            'firstName' : 'raw_string',
            'formattedName' : 'raw_string',
            'headline' : 'raw_text',
            'id' : 'id',
            'industry' : 'raw_enum',
            'isCurrent' : 'raw_boolean',
            'lastName' : 'raw_string',
            'latestPull' : 'raw_datetime',
            'location.country.code' : 'raw_enum',
            'location.name' : 'raw_string',
            'numConnections' : 'raw_integer',
            'pictureUrl' : 'raw_url',
            'publicProfileUrl' : 'raw_url',
            'siteStandardProfileRequest.url': 'raw_url',
            'source.serviceProvider.name' : 'raw_string',
            'startDate.month' : 'raw_integer',
            'startDate.year' : 'raw_integer',
            'summary' : 'raw_text',
            'timestamp' : 'raw_datetime',
            'title' : 'raw_text',
            'updatedAt' : 'raw_datetime',
            'visibility.code' : 'raw_enum'            
        },
        'indexes' : ['timestamp', 'updatedAt']
    },
    'facebook' : {
        'additional_details' : ['picture', 'hometown', 'location', 'significant_other',
                              'locale', 'quotes', 'timezone', 'id', 'first_name', 'verified',
                                'name_format', 'political', 'religion', 'is_verified', 'email', 
                                'short_name', 'is_shared_login', 'birthday', 'link', 'name',
                                'gender', 'updated_time', 'updatedAt', 'createdAt', 'latestPull'],
        'attributes' : ['feed',  'television', 'videos', 'tagged_places', 'movies',
                        'friends', 'photos', 'books', 'music', 'likes', 'family', 'games',
                        'albums', 'video_broadcasts', 'events', 'permissions', 
                        'inspirational_people', 'education', 'sports', 'languages', 'security_settings',
                        'favourite_teams', 'currency', 'favourite_athletes', 'cover',
                        'devices', 'interested_in'],
        'collapse' : ['userEdges', 'userFields'],
        'col_to_type' : {
            'birthday' : 'raw_date',
            'classes' : 'raw_list',
            'concentration' : 'raw_list',
            'createdAt' : 'raw_datetime',
            'created_time' : 'raw_datetime',
            'currency_offset' : 'raw_float',
            'description' : 'raw_text',
            'email' : 'raw_email',
            'end_time' : 'raw_datetime',
            'first_name' : 'raw_string',
            'gender' : 'raw_enum',
            'hardware' : 'raw_string',
            'hometown.id' : 'id',
            'hometown.name' : 'raw_string',
            'id' : 'id',
            'interested_in' : 'raw_enum',
            'is_shared_login' : 'raw_boolean',
            'is_verified' : 'raw_boolean',
            'latestPull' : 'raw_datetime',
            'link' : 'raw_url',
            'locale' : 'raw_enum',
            'location.id' : 'id',
            'location.name' : 'raw_string',
            'message' : 'raw_text',
            'name' : 'raw_string',
            'name_format' : 'raw_string',
            'offset_x' : 'raw_int',
            'offset_y' : 'raw_int',
            'os' : 'raw_string',
            'permission' : 'raw_string',
            'picture.is_silhouette' : 'raw_boolean',
            'picture.url' : 'raw_url',
            'place' : 'raw_object',
            'place.id' : 'id',
            'place.location.city' : 'raw_string',
            'place.location.country' : 'raw_string',
            'place.location.latitude' : 'raw_boolean',
            'place.location.located_in' : 'id',
            'place.location.longitude' : 'raw_float',
            'place.location.state' : 'raw_string',
            'place.location.street' : 'raw_string',
            'place.location.zip' : 'raw_integer',
            'place.name' : 'raw_string',
            'political' : 'raw_string',
            'quotes' : 'raw_text',
            'relationship' : 'raw_enum',
            'religion' : 'raw_string',
            'rsvp_status' : 'raw_enum',
            'school.id' : 'id',
            'school.name' : 'raw_string',
            'secure_browsing.enabled' : 'raw_boolean',
            'short_name' : 'raw_string',
            'source' : 'raw_url',
            'start_time' : 'raw_datetime',
            'status' : 'raw_enum',
            'story' : 'raw_text',
            'timezone' : 'raw_integer',
            'type' : 'raw_enum',
            'updatedAt' : 'raw_datetime',
            'updated_time' : 'raw_datetime',
            'usd_exchange' : 'raw_float',
            'usd_exchange_inverse' : 'raw_float',
            'user_currency' : 'raw_enum',
            'verified' : 'raw_boolean',
            'with' : 'raw_list',
            'year' : 'raw_object',
            'year.id' : 'id',
            'year.name' : 'raw_string'
        },
        'indexes' : ['created_time', 'updatedAt', 'updated_time', 'start_time']
    },
    'fitbit' : {
        'additional_details' : ['displayName', 'id', 'updatedAt', 'createdAt', 'latestPull'],
        'attributes' : ['sleepData', 'fatData', 'bmiData', 'foodData', 'devices', 'weightData', 
                           'activityData', 'heartData', 'lifetimeStats'],
        'col_to_type' : {
            'best.total.distance.date' : 'raw_datetime',
            'best.total.distance.value' : 'raw_float',
            'best.total.floors.date' : 'raw_datetime',
            'best.total.floors.value' : 'raw_integer',
            'best.total.steps.date' : 'raw_datetime',
            'best.total.steps.value' : 'raw_integer',
            'best.tracker.distance.date' : 'raw_datetime',
            'best.tracker.distance.value' : 'raw_float',
            'best.tracker.floors.date' : 'raw_datetime',
            'best.tracker.floors.value' : 'raw_integer',
            'best.tracker.steps.date' : 'raw_datetime',
            'best.tracker.steps.value' : 'raw_integer',
            'createdAt' : 'raw_datetime',
            'date' : 'raw_datetime',
            'displayName' : 'raw_string',
            'foods' : 'raw_list',
            'goals.caloriesOut': 'raw_integer',
            'goals.distance': 'raw_float',
            'goals.floors': 'raw_integer',
            'goals.steps': 'raw_integer',
            'heartRateZones' : 'raw_object',
            'id' : 'id',
            'latestPull' : 'raw_datetime',
            'lifetime.total.caloriesOut' : 'raw_integer',
            'lifetime.total.distance' : 'raw_float',
            'lifetime.total.floors' : 'raw_integer',
            'lifetime.total.steps' : 'raw_integer',
            'lifetime.tracker.caloriesOut' : 'raw_integer',
            'lifetime.tracker.distance' : 'raw_float',
            'lifetime.tracker.floors' : 'raw_integer',
            'lifetime.tracker.steps' : 'raw_integer',
            'restingHeartRate' : 'raw_integer',
            'sleep' : 'raw_list',
            'summary.activityCalories' : 'raw_integer',
            'summary.calories': 'raw_integer',
            'summary.caloriesBMR' : 'raw_integer',
            'summary.caloriesOut': 'raw_integer',
            'summary.carbs': 'raw_integer',
            'summary.distances' : 'raw_list',
            'summary.elevation' : 'raw_float',
            'summary.fairlyActiveMinutes' : 'raw_integer',
            'summary.fat': 'raw_integer',
            'summary.fiber': 'raw_integer',
            'summary.floors' : 'raw_integer',
            'summary.lightlyActiveMinutes' : 'raw_integer',
            'summary.marginalCalories' : 'raw_integer',
            'summary.protein': 'raw_integer',
            'summary.sedentaryMinutes' : 'raw_integer',
            'summary.sodium': 'raw_integer',
            'summary.steps' : 'raw_integer',
            'summary.totalMinutesAsleep' : 'raw_integer',
            'summary.totalSleepRecords' : 'raw_integer',
            'summary.totalTimeInBed' : 'raw_integer',
            'summary.veryActiveMinutes' : 'raw_integer',
            'summary.water': 'raw_integer',
            'updatedAt' : 'raw_datetime',
            'value' : 'raw_float'
        },
        'indexes' : ['date', 'updatedAt']
    },
    'moves' : {
        'attributes' : ['places', 'summaries', 'storyline', 'profile'],
        'col_to_type' : {
            'caloriesAvailable' : 'raw_boolean',
            'createdAt' : 'raw_datetime',
            'currentTimeZone.id' : 'id',
            'currentTimeZone.offset' : 'raw_integer',
            'date' : 'raw_date',
            'firstDate' : 'raw_integer',
            'lastUpdate' : 'raw_datetime',
            'latestPull' : 'raw_datetime',
            'localization.firstWeekDay' : 'raw_integer',
            'localization.language' : 'raw_enum',
            'localization.locale' : 'raw_enum',
            'localization.metric' : 'raw_boolean',
            'platform' : 'raw_enum',
            'segments' : 'raw_list',
            'summary' : 'raw_list',
            'updatedAt' : 'raw_datetime',
            'userId' : 'id'  
        },
        'indexes' : ['lastUpdate', 'updatedAt', 'date'],
        'try_dates' : ['%Y%m%d', '%Y%m%dT%H%M%S%fZ', '%Y-%m-%dT%H:%M:%S.%fZ']
    }
}    


import requests
import os, sys

credentials_file = './credentials.py'

def set_credentials(fil):
	credentials_file = fil

def make_request(credentials_file, endpoint, resource_server = "https://transcend.io", verify=True):
	# Ensure credentials file exists
	if not os.path.exists(credentials_file):
		return("Credentials file does not exist")
	
	#  Read in credentials
	folder, fil = os.path.split(credentials_file)
	sys.path.insert(0, folder)
	credentials = __import__(os.path.splitext(fil)[0])
			
	# the url to make the api request to
	request_url = "%s/api/%s?appId=%s" % (resource_server, endpoint, credentials.CLIENT_ID)
	
	# the header used for authentication of the request
	headers = {"Authorization":"Bearer %s" % credentials.ACCESS_TOKEN}
	
	# send the request
	r = requests.get(request_url, headers = headers, timeout = None, verify=verify)
	
	# Parse the response
	if r.status_code == 200:
		# successful response
		return r.json()
	else:
		return r

def request(endpoint, resource_server = "https://transcend.io", verify=True):
	return make_request(credentials_file, endpoint, resource_server = resource_server, verify=verify)


import pandas as pd
import numpy as np
import datetime
import copy
import json
from config import datascience_format_options

from pandas.io.json import json_normalize

 
def raw_to_data_science(data, serviceURI):
    params = copy.deepcopy(datascience_format_options[serviceURI])
    params['name'] = serviceURI
    params['data'] = data
    if 'indexes' in params:
        del params['indexes']
    return RawToDataScience(**params).to_json()
    

# should be function and the class should be the already formatted object
class RawToDataScience(object):
    def __init__(self, data, name, now = None, dt_str_strip = '%Y-%m-%dT%H:%M:%S.%fZ',
                 date_str_strip = '%m/%d/%Y', collapse = [], col_to_type = {}, attributes = [],
                 getAttributes = {}, additional_details = ['updatedAt', 'createdAt', 'latestPull'],
                 try_dates = None):
        '''
            Initialize a dataset object
        '''
        # the raw data
        self.allData = data
        
        # the time to use as the current time
        self.now = datetime.datetime.now() if now is None else now
        
        # the name of the service
        self.name = name
        
        # the format for datetime strings
        self.dt_str_strip = dt_str_strip
        
        # the format for date strings
        self.date_str_strip = date_str_strip
        
        # the mapping of column name to data type
        self.col_to_type = col_to_type
        
        # the list of attributes one can access
        self.attributes = attributes[:]
        
        # the list of attributes that can be queried at any given time
        self.getAttributes = getAttributes
        
        # the attributes to pull into the details attribute
        self.additional_details = additional_details[:]
        
        # the collumns to initially collapse
        self.collapse = collapse
        
        self.try_dates = try_dates
        
        # pre process the object
        self.pre_process()
    
    def get(self, attr, date = None):
        '''
            Returns a dateframe for the attribute closest to the date in question
        '''
        return self.getAttributeHelper(attr,  date)
    
    def getAll(self, attr):
        '''
            Returns all of the data for the attribute
        '''
        return self.allData[attr]
        
    def pre_process(self):
        for col in self.collapse:
            keys = self.allData[col].keys()
            for key in keys:
                self.allData[key] = self.allData[col][key]
            del self.allData[col]
        
            
        # standard attributes to dataframe
        for attr in self.attributes:
            setattr(self, attr, self.to_df(self.allData[attr], attr))
        
        # bunch together details
        details = pd.Series(self.allData['details'] if 'details' in  self.allData else {})
        for attr in self.additional_details:
            details[attr] = self.allData[attr] if attr in self.allData else ''
        self.details = self.to_df(details, 'details')
        self.attributes.append('details')
    
    def to_json(self):
        response = {
            'attributes' : {},
            'getAttributes' : {}
        }

        for attr in self.attributes:
            response['attributes'][attr] = json.loads(getattr(self, attr).to_json(orient='records'))
        for attr in self.getAttributes:
            response['getAttributes'][attr] = self.getAll(attr)

        return response
            
    def to_string(self):
        '''
            Convert the service object to string and show 
            the attributes
        '''
        # indicate name of the model
        base = 'Transcend Service Model:\n' 
        base += '|- name: %s\n' % self.name
        
        # add raw data
        base += '|- allData #raw json data\n'
        
        # append attributes
        for attr in self.attributes:
              base += '|- %s\n' % attr
        
        # append get attributes
        for attr in self.getAttributes.keys():
              base += '|- get(\'%s\', datetime.datetime.now())\n' % (attr)
        return base
    
    def to_df(self, json_object, name, index = '_id'):
        '''
            Convert a json object to a dataframe
            
            json_object must be of type list, dict or pd.Series
        '''
        obj_type = type(json_object)
        
        # list of objects converted to df
        if obj_type == list:
            df = pd.DataFrame(json_object)
            
            # if not a list of objects, change the column name
            if len(json_object) > 0 and type(json_object[0]) != dict:
                df.columns = [name]
        
        # dict converted to series then to df
        elif obj_type == dict:
            df = pd.Series(json_object).to_frame()
            df.columns = [name]
            df = df.T
        
        # series converted to dataframe
        elif obj_type == pd.Series:
            df = json_object.to_frame().T
        else:
            raise TypeError('Unexpected type for json object %s' % obj_type)
        
        # set the index if it exists
        if index in df:
            df = df.set_index(index)
            
        if df.shape[0] == 0:
            return df
        else:
            df = json_normalize(json.loads(df.to_json(orient='records')))
            
            # convert datetimes and dates to datetime objects
            for col in filter(lambda x : x in self.col_to_type, df.columns):
                col_type = self.col_to_type[col]
                
                if self.try_dates:
                    for form in self.try_dates:
                        try:
                            df[col] = df[col].map(lambda x :  datetime.datetime.strptime(x, form) if pd.notnull(x) else np.nan)
                            break
                        except:
                            pass
                else:
                    if col_type == 'raw_datetime':  
                        df[col] = df[col].map(lambda x :  datetime.datetime.strptime(x, self.dt_str_strip) if pd.notnull(x) else np.nan)
                    elif col_type == 'raw_date':  
                        df[col] = df[col].map(lambda x :  datetime.datetime.strptime(x, self.date_str_strip) if pd.notnull(x) else np.nan)
            return df
    
        

    def __str__(self):
        return self.to_string()
    
    def __repr__(self):
        return self.to_string()

    def getAttributeHelper(self, name, date):
        '''
            Returns the latest pull if date is none,
            else finds the pull closest but after the date
        '''
        rawdata = self.allData[name]
        key = self.getAttributes[name]
        if date is None:
            return self.to_df(rawdata[0][key], name)
        
        df = pd.DataFrame(rawdata)
        df.index = df['timePulled'].map(lambda x :  datetime.datetime.strptime(x, self.dt_str_strip))
        def nearest(items, pivot):
            return min(items, key=lambda x: pivot - x if x < pivot else 1e11)
        nearest_key = nearest(df.index, date)
        if nearest_key > date:
            print('no data from that time')
        else:
            print('closest date %s' % nearest_key)
            return self.to_df(df.ix[nearest_key, key], name)
        
    def get_all_col_keys(self):
        '''
            Get a list of all column headers for the service object
            attributes
        '''
        all_col_keys = set()
        for attr in self.attributes:
            all_col_keys = all_col_keys.union(set(getattr(self, attr).columns))
        for attr in self.getAttributes.keys():
            all_col_keys = all_col_keys.union(set(self.get(attr, self.now).columns))
        return sorted([str(x) for x in list(all_col_keys)])



class DataScience(object):
    def __init__(self, data, name, now = None, dt_str_strip = '%Y-%m-%dT%H:%M:%S.%fZ', date_str_strip = '%m/%d/%Y'):
        '''
            Initialize a dataset object
        '''
        # the raw data
        self.allData = data
        
        # the time to use as the current time
        self.now = datetime.datetime.now() if now is None else now
        
        # the name of the service
        self.name = name
        
        # the format for datetime strings
        self.dt_str_strip = dt_str_strip
        
        # the format for date strings
        self.date_str_strip = date_str_strip
        
        # the mapping of column name to data type
        self.col_to_type = datascience_format_options[name]['col_to_type']
        
        # the list of attributes one can access
        self.attributes = data['attributes'].keys()
        
        # the list of attributes that can be queried at any given time
        self.getAttributes = datascience_format_options[name]['getAttributes'] if 'getAttributes' in datascience_format_options[name] else {}
        
        # pre process the object
        self.pre_process()
    
    def get(self, attr, date = None):
        '''
            Returns a dateframe for the attribute closest to the date in question
        '''
        return self.getAttributeHelper(attr,  date)
    
    def getAll(self, attr):
        '''
            Returns all of the data for the attribute
        '''
        return self.allData['getAttributes'][attr]
        
    def pre_process(self):
            
        # standard attributes to dataframe
        for attr in self.attributes:
            setattr(self, attr, self.to_df(self.allData['attributes'][attr], attr))
        
    def to_json(self):
        response = {
            'attributes' : {},
            'getAttributes' : {}
        }

        for attr in self.attributes:
            response['attributes'][attr] = json.loads(getattr(self, attr).to_json(orient='records') )
        for attr in self.getAttributes:
            response['getAttributes'][attr] = self.getAll(attr)

        return response
            
    def to_string(self):
        '''
            Convert the service object to string and show 
            the attributes
        '''
        # indicate name of the model
        base = 'Transcend Service Model:\n' 
        base += '|- name: %s\n' % self.name
        
        # add raw data
        base += '|- allData #raw json data\n'
        
        # append attributes
        for attr in self.attributes:
              base += '|- %s\n' % attr
        
        # append get attributes
        for attr in self.getAttributes.keys():
              base += '|- get(\'%s\', datetime.datetime.now())\n' % (attr)
        return base
    
    def to_df(self, json_object, name, index = '_id'):
        '''
            Convert a json object to a dataframe
            
            json_object must be of type list, dict or pd.Series
        '''
        df = pd.DataFrame(json_object)

        # set the index if it exists
        if index in df:
            df = df.set_index(index)
            if index == '__index__':
                df.index = pd.to_datetime(df.index, unit='ms')
            
        if df.shape[0] == 0:
            return df
        else:
    
            # convert datetimes and dates to datetime objects
            for col in filter(lambda x : x in self.col_to_type, df.columns):
                col_type = self.col_to_type[col]
                
                if col_type == 'raw_datetime' or col_type == 'raw_date':
                    df[col] = pd.to_datetime(df[col],unit='ms')
            return df
        
    def __str__(self):
        return self.to_string()
    
    def __repr__(self):
        return self.to_string()

    def getAttributeHelper(self, name, date):
        '''
            Returns the latest pull if date is none,
            else finds the pull closest but after the date
        '''
        rawdata = self.allData['getAttributes'][name]
        key = self.getAttributes[name]
        if date is None:
            return self.to_df(rawdata[0][key], name)
        
        df = pd.DataFrame(rawdata)
        df.index = df['timePulled'].map(lambda x :  datetime.datetime.strptime(x, self.dt_str_strip))
        def nearest(items, pivot):
            return min(items, key=lambda x: pivot - x if x < pivot else 1e11)
        nearest_key = nearest(df.index, date)
        if nearest_key > date:
            print('no data from that time')
        else:
            print('closest date %s' % nearest_key)
            return self.to_df(df.ix[nearest_key, key], name)
        
    def get_all_col_keys(self):
        '''
            Get a list of all column headers for the service object
            attributes
        '''
        all_col_keys = set()
        for attr in self.attributes:
            all_col_keys = all_col_keys.union(set(getattr(self, attr).columns))
        for attr in self.getAttributes.keys():
            all_col_keys = all_col_keys.union(set(self.get(attr, self.now).columns))
        return sorted([str(x) for x in list(all_col_keys)])


#python setup.py sdist upload
from setuptools import setup
setup(
	name='transcend',    # This is the name of your PyPI-package.
	version='1.3.17',                          # Update the version number for new releases
	author='Michael Farrell',
	author_email='mike@transcend.io',
	url='https://github.com/transcend-inc/transcend-sdk-python',
	packages=['transcend', 'transcend/formats'] 
)